#include "bmploader.h"
#include "exception.h"

struct BmpFileHdr
{
	WORD bfType;		// 'BM' sizeof == 14
	DWORD bfSize;		// file size
	WORD bfReserved1;
	WORD bfReserved2;
	DWORD bfOffBits;	//offset to picture
};

struct BmpInfoHdr
{
	DWORD biSize;			// sizeof(mtBmpInfoHdr) == 40
	DWORD biWidth;
	DWORD biHeight;
	WORD biPlanes;			// == 1
	WORD biBitCount;		// == 1 or 4 or 8 or 24 or 32
	DWORD biCompression;	// 0 - no compressed, 1 - RLE8, 2 - RLE4
	DWORD biSizeImage;		// sizeof(picture)
	DWORD biXPelsPerMeter;
	DWORD biYPelsPerMeter;
	DWORD biClrUsed;		// hmColors used in picture
	DWORD biClrImportant;	// hm important colors
};

void LoadBMP(File* fl, int* width, int* height, DWORD** data, bool ignoreAlpha)
{
	BmpFileHdr bmpFileHdr;
	BmpInfoHdr bmpInfoHdr;

	bmpFileHdr.bfType = fl->GetWORD();
	bmpFileHdr.bfSize = fl->GetDWORD();
	bmpFileHdr.bfReserved1 = fl->GetWORD();
	bmpFileHdr.bfReserved2 = fl->GetWORD();
	bmpFileHdr.bfOffBits = fl->GetDWORD();

	if (bmpFileHdr.bfType != 0x4D42) throw Exception("Invalid BMP header");

	bmpInfoHdr.biSize = fl->GetDWORD();
	bmpInfoHdr.biWidth = fl->GetDWORD();
	bmpInfoHdr.biHeight = fl->GetDWORD();
	bmpInfoHdr.biPlanes = fl->GetWORD();
	bmpInfoHdr.biBitCount = fl->GetWORD();
	bmpInfoHdr.biCompression = fl->GetDWORD();
	bmpInfoHdr.biSizeImage = fl->GetDWORD();
	bmpInfoHdr.biXPelsPerMeter = fl->GetDWORD();
	bmpInfoHdr.biYPelsPerMeter = fl->GetDWORD();
	bmpInfoHdr.biClrUsed = fl->GetDWORD();
	bmpInfoHdr.biClrImportant = fl->GetDWORD();

	if (bmpInfoHdr.biBitCount!=24 && bmpInfoHdr.biBitCount!=32) throw Exception("%d-bit bmp's are not supported", bmpInfoHdr.biBitCount);

	fl->SetFilePointer(bmpFileHdr.bfOffBits);

	*width = bmpInfoHdr.biWidth;
	*height = bmpInfoHdr.biHeight;
	*data = new DWORD[(*width) * (*height)];

	if (bmpInfoHdr.biBitCount == 24)
	{
		int k = (bmpInfoHdr.biWidth * 3) % 4;
		if (k != 0) k = 4 - k;

		for (int i = bmpInfoHdr.biHeight-1; i >= 0; i--)
		{
			for (int j = 0; (unsigned)j < bmpInfoHdr.biWidth; j++)
			{
				int b = fl->GetBYTE();
				int g = fl->GetBYTE();
				int r = fl->GetBYTE();
				(*data)[i * (*width) + j] = DRGB(r, g, b);
			}

			for (int j = 0; j < k; j++) fl->GetBYTE();
		}
	}
	else
	{
		for (int i = bmpInfoHdr.biHeight-1; i >= 0; i--)
		{
			for (int j = 0; (unsigned)j < bmpInfoHdr.biWidth; j++)
			{
				int b = fl->GetBYTE();
				int g = fl->GetBYTE();
				int r = fl->GetBYTE();
				int a = fl->GetBYTE();
				(*data)[i * (*width) + j] = DRGBA(r, g, b, (ignoreAlpha ? 0 : a));
			}
		}
	}
}
