#include "channel.h"
#include "global.h"

MusicChannel::MusicChannel()
{
	loaded = false;
	volume = 0x100;

	bufferPtr = 0;
	bufferSize = 0;

	buffer = new BYTE[MUSIC_CHANNEL_BUFFER_SIZE];
	mux = SDL_CreateMutex();
}

MusicChannel::~MusicChannel()
{
	Unload();
	delete[] buffer;
	SDL_DestroyMutex(mux);
}

MusicChannel::MusicChannel(const MusicChannel& f)
{
	Error("Copy MusicChannel object not allowed (construct)");
}

MusicChannel& MusicChannel::operator=(const MusicChannel& f)
{
	Error("Copy MusicChannel object not allowed (operator=)");
}

void MusicChannel::FillBuffer(BYTE *data, int count)
{
	if (count <= 0) return;
	if (count & 3) Error("[MusicChannel::FillBuffer] Count must divides by 4 without reminder");
	if (!enableSound) return;

	if (!loaded)
	{
		while (count--) *(data++) = 0;
		return;
	}

	for (;;)
	{
		if (bufferPtr < bufferSize)
		{
			unsigned int sz = bufferSize - bufferPtr;
			if (sz > count) sz = count;

			BYTE* ptr = &buffer[bufferPtr];

			count -= sz;
			bufferPtr += sz;

			sz >>= 1;

			while (sz--)
			{
				int v = (Sint16)((ptr[1] << 8) | (*ptr));

				v = v * volume / 0x100;

				*(data++) = (v & 0xFF);
				*(data++) = (v >> 8);

				ptr += 2;
			}

			if (count <= 0) break;
		}
		else
		{
			SDL_mutexP(mux);
			bufferSize = ov_read(&vf, (char*)buffer, MUSIC_CHANNEL_BUFFER_SIZE, 0, 2, 1, &currentSection);
			SDL_mutexV(mux);

			bufferPtr = 0;

			if (bufferSize <= 0)
			{
				while (count--) { *(data++) = 0; }

				if (bufferSize == 0)
				{
					if (loop) Load(name, loop);
					else Unload();
				}

				break;
			}
		}
	}
}

void MusicChannel::Unload(void)
{
	if (!enableSound) return;
	if (!loaded) return;

	SDL_mutexP(mux);

	ov_clear(&vf);
	fl.Close();

	loaded = false;

	SDL_mutexV(mux);
}

size_t ReadOggData(void *ptr, size_t size, size_t nmemb, void *datasource)
{
	File* fl = (File *)datasource;
	return fl->ReadBytes((BYTE*)ptr, size * nmemb);
}

void MusicChannel::Load(string name, bool loop)
{
	if (!enableSound) return;
	if (loaded) Unload();

	fl.Read(name.c_str());

	ov_callbacks cb;
	cb.read_func = ReadOggData;
	cb.seek_func = NULL;
	cb.close_func = NULL;
	cb.tell_func = NULL;

	if (ov_open_callbacks(&fl, &vf, NULL, 0, cb) < 0) Error("Input does not appear to be an Ogg bitstream");

	this->name = name;
	this->loop = loop;

	currentSection = 0;
	loaded = true;
}

void MusicChannel::SetVolume(int volume)
{
	if (volume<0 || volume>0x100) throw "Volume must be between 0 and 0x100";
	this->volume = volume;
}

int MusicChannel::GetVolume(void)
{
	return volume;
}

bool MusicChannel::IsLoaded(void)
{
	return loaded;
}
