#include "failsystem.h"
#include "failhandle.h"

#include "defines.h"
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifndef FAIL_CODE
	#define FAIL_CODE "UWB4TP"
#endif

#ifndef FAIL_ARCHIVE_NAME
	#define FAIL_ARCHIVE_NAME "data.fail"
#endif

#ifndef O_BINARY
	#define O_BINARY 0
#endif

FailSystem::FailSystem()
{
	alreadyInit = false;
	encSz = strlen(FAIL_CODE);
}

FailSystem::~FailSystem()
{
}

int FailSystem::NormalizeEncOff(int encOff)
{
	if (!encSz) return 0;
	return (encOff < 0 ? ((encOff % encSz) + encSz) : (encOff % encSz));
}

int FailSystem::Decrypt(int encOff, void* data, unsigned int size)
{
	if (!encSz) return encOff;

	unsigned char* rdata = (unsigned char*)data;

	while (size)
	{
		*(rdata++) = (*rdata) ^ FAIL_CODE[encOff];
		encOff = (encOff + 1) % encSz;
		size--;
	}

	return encOff;
}

void FailSystem::InitFS(const char* arcName)
{
	files.clear();
	currentArcName = string(arcName);

	int handle = open(arcName, O_BINARY | O_RDONLY);
	if (handle == -1) throw Exception("Fail to open \"%s\"", arcName);

	char idbuf[] = "12345";
	read(handle, idbuf, 5);
	if (strcmp(idbuf, "FAIL!")) throw Exception("Fail to identify \"%s\"", arcName);

	int encOff = 0;

	unsigned char dirSizeBuf[2];
	if (read(handle, dirSizeBuf, 2) != 2) throw Exception("Fail to read \"%s\" (A)", arcName);
	encOff = Decrypt(encOff, dirSizeBuf, 2);
	unsigned short dirSize = (unsigned short)dirSizeBuf[0] + 0x100 * (unsigned short)dirSizeBuf[1];

	unsigned char* buf = new unsigned char[dirSize];
	if (read(handle, buf, dirSize) != dirSize) throw Exception("Fail to read \"%s\" (B)", arcName);
	encOff = Decrypt(encOff, buf, dirSize);

	int p = 0;
	long lastOffset = lseek(handle, 0L, SEEK_CUR);

	char namebuf[0x100];

	while (p < dirSize)
	{
		if (p+5 > dirSize) throw Exception("Fail to read \"%s\" (C)", arcName);

		long size = (long)buf[p] + 0x100L*(long)buf[p+1] + 0x10000L*(long)buf[p+2] + 0x1000000L*(long)buf[p+3];

		char *np = namebuf;
		p += 4;

		while (buf[p]) *(np++) = buf[p++];

		*np = 0;
		p++;

		FailEntry entry;
		entry.name = string(namebuf);
		entry.offset = lastOffset;
		entry.size = size;
		entry.encOff = encOff;
		files.push_back(entry);

		lastOffset += size;
		encOff = NormalizeEncOff(encOff + size);
	}

	close(handle);
	delete[] buf;
}

FileHandle* FailSystem::GetHandle(const char* filename, int mode)
{
	if (mode != FS_READ) return NULL;

	if (!alreadyInit)
	{
		InitFS(FAIL_ARCHIVE_NAME);
		alreadyInit = true;
	}

	string name = string(filename);
	list<FailEntry>::iterator it;

	for (it = files.begin(); it != files.end(); it++)
	{
		if (it->name == name)
		{
			int hndl = open(currentArcName.c_str(), O_BINARY | O_RDONLY);
			if (hndl == -1) throw Exception("Fail to open \"%s\"", currentArcName.c_str());

			lseek(hndl, it->offset, SEEK_SET);

			FailHandle* fh = new FailHandle(this);
			fh->handle = hndl;
			fh->offset = it->offset;
			fh->size = it->size;
			fh->encOff = it->encOff;
			return fh;
		}
	}

	return NULL;
}

void FailSystem::CloseHandle(FileHandle* fh)
{
	close(((FailHandle*)fh)->handle);
	delete fh;
}
