#ifndef __FILE_H__
#define __FILE_H__

#include "defines.h"
#include "filesystem.h"
#include <SDL.h>

#define ACC_NONE 0
#define ACC_READ 1
#define ACC_WRITE 2
#define ACC_APPEND 3

#define FILECACHE_SIZE 0x1000

class File
{
	public:

	File();
	File(FileSystem* fs);
	virtual ~File();

	File(const File& f);
	File& operator=(const File& f);

	void Read(const char *filename);
	void Write(const char *filename);
	void Append(const char *filename);
	void Close(void);

	void WriteBytes(BYTE* buf, unsigned int size);
	void PrintF(const char* fmt, ...);
	void PutBYTE(BYTE b);
	void PutWORD(WORD w);
	void PutDWORD(DWORD d);
	void PutC(int c);

	unsigned int ReadBytes(BYTE *buf, unsigned int size);
	void GetS(char *buf, unsigned int size);
	BYTE GetBYTE(void);
	WORD GetWORD(void);
	DWORD GetDWORD(void);
	int GetC(void);

	bool Eof(void);
	void UnGetBYTE(BYTE b);
	void UnGetC(int c);

	long GetCurrentFileSize(void);
	void SetFilePointer(long fp);

	long FileSize(const char *filename);
	bool FileExists(const char *filename);

	bool IsOpened(void);

	SDL_RWops* AllocSdlRwOps(void);
	void FreeSdlRwOps(SDL_RWops* rwo);

	static int SdlRwOpsSeekFunc(SDL_RWops* context, int offset, int whence);
	static int SdlRwOpsReadFunc(SDL_RWops* context, void* ptr, int size, int maxnum);
	static int SdlRwOpsWriteFunc(SDL_RWops* context, const void* ptr, int size, int num);
	static int SdlRwOpsCloseFunc(SDL_RWops* context);

	protected:

	void TryNClose(void);
	void EnsureAccMode(int accMode);

	BYTE buffer[FILECACHE_SIZE];

	FileSystem* fs;
	FileHandle* handle;

	int len, accMode, un_ch;
	long readSize, currentFileSize;
	bool eof, un_eof;
};

#endif
