#include "game.h"
#include "global.h"
#include "scenario.h"
#include "cache.h"
#include "exception.h"
#include "font.h"
#include "sound.h"
#include "animation.h"

#include <vector>
using namespace std;

struct MenuItem
{
	string text;
	string label;
	int lineNum;
};

#define FONT_TEXT 0
#define FONT_NAMES 1
#define FONT_INFO 2
#define FONT_MENU 3
#define FONT_MENU_SELECTED 4
#define FONT_CHAPTER 5
#define FONT_LAST (5+1)

Font fonts[FONT_LAST];
Cache cache;
Scenario scen;
Picture back, charLeft, charCenter, charRight, menuBg, chapterBg;
Animation menuWait, textWait;
string charName(""), currentText(""), currentInfo("");
int currentDelay = 0;
vector<MenuItem> currentMenu;
bool showMenu = false;
bool lmbPressed = false;
int ntick = 0;
bool hiSpeedForward = false;

int lastchars, lastCharX, lastCharY;

int textOffsetX = 0;
int textOffsetY = 0;
int textOffsetR = 0;

int GetStringWidth(const char* str, int fontStyle)
{
	int x = 0;

	while (*str)
	{
		x += fonts[fontStyle].Width(*str);
		str++;
	}

	return x;
}

void PrintString(int x, int y, const char* str, int fontStyle, int fontStyleWidthAdd)
{
	int sx = x;

	while (*str)
	{
		if (*str == '\n')
		{
			y += fonts[fontStyle].Height();
			x = sx;
		}
		else
		{
			fonts[fontStyle].Blit(screen, x, y, *str);
			x += fonts[fontStyleWidthAdd].Width(*str);			
		}

		str++;
	}
}

void PrintString(int x, int y, const char* str, int fontStyle)
{
	PrintString(x, y, str, fontStyle, fontStyle);
}

void PrintWords(int x, int y, const char* str, int fontStyle, int fontStyleWidthAdd, int fromchar, int tochar)
{
	int sx = x;
	int cnt = 0;
	bool checkWord = true;

	int h = fonts[fontStyle].Height();

	while (*str)
	{
		if (*str == '\n')
		{
			y += h;
			x = sx;
			checkWord = true;
		}
		else
		{
			if (*str == ' ') checkWord = true;
			else
			if (checkWord)
			{
				const char* ptr = str;
				int wdt = 0;

				while (*ptr && *ptr!=' ') wdt += fonts[fontStyleWidthAdd].Width(*(ptr++));

				if (x + wdt > screen->w - textOffsetR)
				{
					y += h;
					x = sx;
				}

				checkWord = false;
			}

			int w = fonts[fontStyleWidthAdd].Width(*str);

			if (cnt >= fromchar)
			{
				fonts[fontStyle].Blit(screen, x, y, *str);
				updateList.push_back(DefRect(x, y, w, h));
			}

			x += w;

			cnt++;
			if (cnt > tochar) break;
		}

		str++;
	}

	x += 4;

	if (x + textWait.Width() > screen->w - textOffsetR)
	{
		y += h;
		x = sx;
	}

	lastCharX = x;
	lastCharY = y;
}

void ToScreenBuffer(void)
{
	SDL_Rect r = DefRect(0, screen->h - menuBg.Height(), screen->w, menuBg.Height());
	SDL_BlitSurface(screen, &r, screenBuffer, &r);
}

void FromScreenBuffer(void)
{
	SDL_Rect r = DefRect(0, screen->h - menuBg.Height(), screen->w, menuBg.Height());
	SDL_BlitSurface(screenBuffer, &r, screen, &r);
}

void UpdatePicts(void)
{
	Clear(0);
	back.Blit(screen, 0, 0);
	charLeft.Blit(screen, 0, 0);
	charCenter.Blit(screen, (screen->w - charCenter.Width()) / 2, 0);
	charRight.Blit(screen, screen->w - charRight.Width(), 0);

	ToScreenBuffer();

	int tx = textOffsetX;
	int mt = screen->h - menuBg.Height();

	menuBg.Blit(screen, 0, mt);
	updateAll = true;
}

bool DrawText(int nchars)
{
	int tx = textOffsetX;
	int ty = screen->h - menuBg.Height() + textOffsetY;

	if (!charName.empty())
	{
		string cname = charName + string(": ");
		PrintString(tx, ty, cname.c_str(), FONT_NAMES, FONT_TEXT);
		tx += GetStringWidth(cname.c_str(), FONT_TEXT);
	}

	PrintWords(tx, ty, currentText.c_str(), FONT_TEXT, FONT_TEXT, lastchars, nchars);
	lastchars = nchars + 1;

	return (nchars > currentText.length());
}

void DrawTextWaitIndicator(void)
{
	int x = lastCharX;
	int y = lastCharY + 1;

	SDL_Rect s = DefRect(x, y - (screen->h - menuBg.Height()), textWait.Width(), textWait.Height());
	SDL_Rect d = DefRect(x, y, textWait.Width(), textWait.Height());

	SDL_BlitSurface(screenBuffer, &d, screen, &d);
	menuBg.Blit(screen, &s, &d);

	textWait.Blit(screen, x, y, SDL_GetTicks());
	updateList.push_back(d);
}

int DrawMenu(void)
{
	int res = -1;

	FromScreenBuffer();
	menuBg.Blit(screen, 0, screen->h - menuBg.Height());

	for (int i = 0; i < currentMenu.size(); i++)
	{
		int x = textOffsetX;
		int y = screen->h - menuBg.Height() + textOffsetY + i * fonts[FONT_MENU].Height();
		int lx = x + GetStringWidth(currentMenu[i].text.c_str(), FONT_MENU) + menuWait.Width() + 8;
		int ly = y + fonts[FONT_MENU].Height();

		updateList.push_back(DefRect(x, y, lx-x, ly-y));
		
		menuWait.Blit(screen, x, y, tick);
		x += menuWait.Width() + 8;

		if (mouseX>=x && mouseY>=y && mouseX<lx && mouseY<ly)
		{
			PrintString(x, y, currentMenu[i].text.c_str(), FONT_MENU_SELECTED, FONT_MENU);
			res = i;
		}
		else PrintString(x, y, currentMenu[i].text.c_str(), FONT_MENU);
	}

	cursorShape = (res == -1 ? CURSOR_ARROW : CURSOR_HAND);
	return res;
}

void ShowPic(void)
{
	SDL_Rect r = DefRect(0, screen->h - menuBg.Height(), screen->w, menuBg.Height());

	if (!FrameBegin()) return;
	SDL_BlitSurface(screen, &r, screenSecondaryBuffer, &r);
	FromScreenBuffer();
	updateAll = true;
	FrameEnd();

	do
	{
		if (!FrameBegin()) continue;
		FrameEnd();
		
		if (key == SDLK_TAB)
		{
			KeyPressed();
			break;
		}
	} while (isActive);

	if (!FrameBegin()) return;
	SDL_BlitSurface(screenSecondaryBuffer, &r, screen, &r);
	updateAll = true;
	FrameEnd();
}

void DrawInfo(void)
{
	int tx = textOffsetX;
	int ty = screen->h - menuBg.Height() + textOffsetY;

	PrintString(tx, ty, currentInfo.c_str(), FONT_INFO, FONT_TEXT);
}

void GameLoop(void)
{
	bool needUpdate = true;
	bool hasDelay = (currentDelay > 0);
	int selectedItem = -1;
	int stick = SDL_GetTicks();

	if (showMenu)
	{
		hasDelay = false;
		currentText = string("");
		currentInfo = string("");
	}
	else
	if (!currentInfo.empty()) currentText = string("");

	lastchars = 0;
	bool allTextPrinted = currentText.empty();
	bool firstWait = true;

	do
	{
		if (!FrameBegin()) continue;
		bool breakIt = false;

		if (key==SDLK_RALT && !showMenu)
		{
			if (!hiSpeedForward)
			{
				Clear(0x000000);
				updateAll = true;
				hiSpeedForward = true;
			}

			breakIt = true;
		}
		else
		{
			hiSpeedForward = false;
			if (needUpdate) UpdatePicts();

			if (!currentText.empty())
			{
				if (allTextPrinted)
				{
					if (firstWait)
					{
						textWait.SetStartTick(tick);
						DrawText(9999);
						firstWait = false;
					}

					DrawTextWaitIndicator();
				}
				else { allTextPrinted = DrawText((SDL_GetTicks() - stick) / 10); }
			}

			if (!currentInfo.empty()) {
				if (needUpdate) {
					DrawInfo();
				}
			}

			if (showMenu && currentMenu.size()>0) { selectedItem = DrawMenu(); }

			needUpdate = false;
			if (!mouseLB) lmbPressed = false;

			if (key==SDLK_LALT && !showMenu)
			{
				DrawText(9999);
				breakIt = true;
			}

			if (mouseLB && !lmbPressed && tick>ntick && (!showMenu || selectedItem>=0))
			{
				lmbPressed = true;
				ntick = tick + 100;

				if (!allTextPrinted)
				{
					DrawText(9999);
					allTextPrinted = true;
				}
				else breakIt = true;
			}
		}

		FrameEnd(!hiSpeedForward);

		if (!breakIt && key==SDLK_TAB)
		{
			KeyPressed();
			ShowPic();
		}

		if (breakIt) break;
		if (hasDelay && tick > stick+currentDelay) break;
	} while (isActive);

	cursorShape = CURSOR_ARROW;

	if (isActive)
	{
		currentDelay = 0;
		currentText = string("");
		currentInfo = string("");
		showMenu = false;

		if (selectedItem >= 0)
		{
			scen.GotoLabel(currentMenu[selectedItem].label, currentMenu[selectedItem].lineNum);
			currentMenu.clear();
		}
	}
}

void ShowChapter(const char *text)
{
	int x = (chapterBg.Width() - GetStringWidth(text, FONT_CHAPTER)) / 2;
	int y = (chapterBg.Height() - fonts[FONT_CHAPTER].Height()) / 2;

	bool drawIt = true;

	do
	{
		if (!FrameBegin()) continue;

		if (drawIt)
		{
			chapterBg.Blit(screen, 0, 0);
			PrintString(x, y, text, FONT_CHAPTER, FONT_CHAPTER);
			updateAll = true;
		}

		FrameEnd();

		if (!mouseLB) lmbPressed = false;

		if (mouseLB && !lmbPressed && tick>ntick)
		{
			lmbPressed = true;
			ntick = tick + 100;
			break;
		}
	} while (isActive);
}

void FunctionChapter(hash_map<string, Scenario::Parameter> params, Scenario* scen, int lineNum)
{
	if (params["text"].type == PARAM_NONE) throw Exception("No chapter text in %s at line %d", scen->scriptPath.c_str(), lineNum);
	if (params["text"].type =! PARAM_STRING) throw Exception("Invalid chapter param type in %s at line %d", scen->scriptPath.c_str(), lineNum);

	ShowChapter(params["text"].strValue.c_str());
}

void FunctionBg(hash_map<string, Scenario::Parameter> params, Scenario* scen, int lineNum)
{
	if (params["name"].type == PARAM_NONE) throw Exception("No back name in %s at line %d", scen->scriptPath.c_str(), lineNum);

	if (params["name"].type == PARAM_STRING)
	{
		back = *cache.LoadPicture(string("gfx/") + params["name"].strValue + string(".png"));
	}
	else
	if (params["name"].type == PARAM_SYMBOL)
	{
		if (params["name"].strValue != string("none")) throw Exception("Only :none allowed in %s at line %d", scen->scriptPath.c_str(), lineNum);
		back.Clear();
	}
	else throw Exception("Invalid back name in %s at line %d", scen->scriptPath.c_str(), lineNum);

	if (params["clear"].type != PARAM_NONE)
	{
		if (params["clear"].type != PARAM_SYMBOL) throw Exception("Invalid clear type in %s at line %d", scen->scriptPath.c_str(), lineNum);
		if (params["clear"].strValue!="no" && params["clear"].strValue!="yes") throw Exception("Invalid clear value in %s at line %d", scen->scriptPath.c_str(), lineNum);

		if (params["clear"].strValue == string("no")) return;
	}

	charLeft.Clear();
	charCenter.Clear();
	charRight.Clear();
	currentText = "";
}

void FunctionChar(hash_map<string, Scenario::Parameter> params, Scenario* scen, int lineNum)
{
	if (params["name"].type == PARAM_NONE) throw Exception("No char name in %s at line %d", scen->scriptPath.c_str(), lineNum);

	int al = 0;

	if (params["align"].type != PARAM_NONE)
	{
		if (params["align"].type != PARAM_SYMBOL) throw Exception("Invalid align type in %s at line %d", scen->scriptPath.c_str(), lineNum);

		if (params["align"].strValue == string("left")) al = -1;
		else
		if (params["align"].strValue == string("center")) al = 0;
		else
		if (params["align"].strValue == string("right")) al = 1;
		else throw Exception("Invalid align value in %s at line %d", scen->scriptPath.c_str(), lineNum);
	}

	if (params["name"].type == PARAM_STRING)
	{
		switch (al)
		{
			case -1:
				charLeft = *cache.LoadPicture(string("gfx/") + params["name"].strValue + string(".png"));
				break;

			case 0:
				charCenter = *cache.LoadPicture(string("gfx/") + params["name"].strValue + string(".png"));
				break;

			case 1:
				charRight = *cache.LoadPicture(string("gfx/") + params["name"].strValue + string(".png"));
				break;
		}
	}
	else
	if (params["name"].type == PARAM_SYMBOL)
	{
		if (params["name"].strValue != string("none")) throw Exception("Only :none allowed in %s at line %d", scen->scriptPath.c_str(), lineNum);

		switch (al)
		{
			case -1:
				charLeft.Clear();
				break;

			case 0:
				charCenter.Clear();
				break;

			case 1:
				charRight.Clear();
				break;
		}
	}
}

void FunctionInfo(hash_map<string, Scenario::Parameter> params, Scenario* scen, int lineNum)
{
	if (params["text"].type == PARAM_NONE) throw Exception("No info text in %s at line %d", scen->scriptPath.c_str(), lineNum);
	if (params["text"].type =! PARAM_STRING) throw Exception("Invalid info param type in %s at line %d", scen->scriptPath.c_str(), lineNum);

	currentInfo = params["text"].strValue;
	GameLoop();
}

void FunctionSay(hash_map<string, Scenario::Parameter> params, Scenario* scen, int lineNum)
{
	if (params["name"].type != PARAM_NONE)
	{
		if (params["name"].type != PARAM_STRING) throw Exception("Invalid name type in %s at line %d", scen->scriptPath.c_str(), lineNum);
		charName = params["name"].strValue;
	}
	else charName = "";

	if (params["sound"].type != PARAM_NONE)
	{
		if (params["sound"].type != PARAM_STRING) throw Exception("Invalid sound type in %s at line %d", scen->scriptPath.c_str(), lineNum);

		if (!params["sound"].strValue.empty())
		{
			channels[2].Load(string("mfx/voc/") + params["sound"].strValue + string(".ogg"), false);
		}
		else
		{
			channels[2].Unload();
		}
	}
	else
	{
		channels[2].Unload();
	}

	if (params["text"].type == PARAM_NONE) throw Exception("Text required in %s at line %d", scen->scriptPath.c_str(), lineNum);
	if (params["text"].type != PARAM_STRING) throw Exception("Invalid text type in %s at line %d", scen->scriptPath.c_str(), lineNum);

	currentText = params["text"].strValue;
	GameLoop();
}

void FunctionItem(hash_map<string, Scenario::Parameter> params, Scenario* scen, int lineNum)
{
	if (params["text"].type == PARAM_NONE) throw Exception("Text required in %s at line %d", scen->scriptPath.c_str(), lineNum);
	if (params["text"].type != PARAM_STRING) throw Exception("Invalid text type in %s at line %d", scen->scriptPath.c_str(), lineNum);

	if (params["label"].type == PARAM_NONE) throw Exception("Label required in %s at line %d", scen->scriptPath.c_str(), lineNum);
	if (params["label"].type != PARAM_SYMBOL) throw Exception("Invalid label type in %s at line %d", scen->scriptPath.c_str(), lineNum);

	MenuItem mi;
	mi.text = params["text"].strValue;
	mi.label = params["label"].strValue;
	mi.lineNum = lineNum;
	currentMenu.push_back(mi);
}

void FunctionMenu(hash_map<string, Scenario::Parameter> params, Scenario* scen, int lineNum)
{
	if (params["action"].type == PARAM_NONE) throw Exception("Action required in %s at line %d", scen->scriptPath.c_str(), lineNum);
	if (params["action"].type != PARAM_SYMBOL) throw Exception("Invalid action type in %s at line %d", scen->scriptPath.c_str(), lineNum);

	if (params["action"].strValue == string("run")) showMenu = true;
	GameLoop();
}

void FunctionDelay(hash_map<string, Scenario::Parameter> params, Scenario* scen, int lineNum)
{
	if (params["value"].type == PARAM_NONE) throw Exception("Value required in %s at line %d", scen->scriptPath.c_str(), lineNum);
	if (params["value"].type != PARAM_NUMBER) throw Exception("Invalid value type in %s at line %d", scen->scriptPath.c_str(), lineNum);

	currentDelay = params["value"].numValue * 100;
	GameLoop();
}

void FunctionGame(hash_map<string, Scenario::Parameter> params, Scenario* scen, int lineNum)
{
	if (params["action"].type == PARAM_NONE) throw Exception("Action required in %s at line %d", scen->scriptPath.c_str(), lineNum);
	if (params["action"].type != PARAM_SYMBOL) throw Exception("Invalid action type in %s at line %d", scen->scriptPath.c_str(), lineNum);

	if (params["action"].strValue == string("end")) isActive = false;
}

void FunctionPreload(hash_map<string, Scenario::Parameter> params, Scenario* scen, int lineNum)
{
	if (params["name"].type == PARAM_NONE) throw Exception("Name required in %s at line %d", scen->scriptPath.c_str(), lineNum);
	if (params["name"].type != PARAM_STRING) throw Exception("Invalid name type in %s at line %d", scen->scriptPath.c_str(), lineNum);

	// cursorShape = CURSOR_BUSY;
	back = *cache.LoadPicture(string("gfx/") + params["name"].strValue + string(".png"));
	// cursorShape = CURSOR_ARROW;
}

void DefineFunction(void (* func)(hash_map<string, Scenario::Parameter>, Scenario*, int), const char* name)
{
	Scenario::EngineFunc efunc;
	efunc.func = func;
	scen.DefineFunc(name, efunc);
}

void DefineFunction(void (* func)(hash_map<string, Scenario::Parameter>, Scenario*, int), const char* name, const char* param1)
{
	Scenario::EngineFunc efunc;
	efunc.func = func;
	efunc.paramNames.push_back(param1);
	scen.DefineFunc(name, efunc);
}

void DefineFunction(void (* func)(hash_map<string, Scenario::Parameter>, Scenario*, int), const char* name, const char* param1, const char* param2)
{
	Scenario::EngineFunc efunc;
	efunc.func = func;
	efunc.paramNames.push_back(param1);
	efunc.paramNames.push_back(param2);
	scen.DefineFunc(name, efunc);
}

void DefineFunction(void (* func)(hash_map<string, Scenario::Parameter>, Scenario*, int), const char* name, const char* param1, const char* param2, const char* param3)
{
	Scenario::EngineFunc efunc;
	efunc.func = func;
	efunc.paramNames.push_back(param1);
	efunc.paramNames.push_back(param2);
	efunc.paramNames.push_back(param3);
	scen.DefineFunc(name, efunc);
}

void InitFunctions(void)
{
	DefineFunction(FunctionChapter, "chapter", "text");
	DefineFunction(FunctionBg, "bg", "name", "clear");
	DefineFunction(FunctionChar, "char", "name", "align");
	DefineFunction(FunctionInfo, "info", "text");
	DefineFunction(FunctionSay, "say", "name", "sound", "text");
	DefineFunction(FunctionItem, "item", "text", "label");
	DefineFunction(FunctionMenu, "menu", "action");
	DefineFunction(FunctionDelay, "delay", "value");
	DefineFunction(FunctionGame, "game", "action");
	DefineFunction(FunctionPreload, "preload", "name");
}

void LoadFont(int fontType, int fontSize, string prefix)
{
	int style = 0;

	if (settings.Get((prefix + string(".Bold")).c_str(), 0)) style |= TTF_STYLE_BOLD;
	if (settings.Get((prefix + string(".Italic")).c_str(), 0)) style |= TTF_STYLE_ITALIC;

	fonts[fontType].Load("fnt/font.ttf", fontSize, style);
	fonts[fontType].SetColor(settings.Get((prefix + string(".Color")).c_str(), 0x000000));
}

void RunGame(void)
{
	InitFunctions();

	textOffsetX = settings.Get("Offset.TextX", 0);
	textOffsetY = settings.Get("Offset.TextY", 0);
	textOffsetR = settings.Get("Offset.TextR", 0);

	unsigned int textSize = settings.Get("Font.TextSize", 16);
	unsigned int menuSize = settings.Get("Font.MenuSize", 16);
	unsigned int chapterSize = settings.Get("Font.ChapterSize", 32);

	LoadFont(FONT_TEXT, textSize, string("Font.Text"));
	LoadFont(FONT_NAMES, textSize, string("Font.Names"));
	LoadFont(FONT_INFO, textSize, string("Font.Info"));
	LoadFont(FONT_MENU, menuSize, string("Font.Menu"));
	LoadFont(FONT_MENU_SELECTED, menuSize, string("Font.Menu.Selected"));
	LoadFont(FONT_CHAPTER, chapterSize, string("Font.Chapter"));

	menuBg.Load("system/menu.png");
	chapterBg.Load("system/chapter.png");

	textWait.Load("system/textwait.anim");
	menuWait.Load("system/menuwait.anim");

	scen.Load("scenario/main.script");
	scen.Init();

	while (isActive) {
		scen.Step();
	}
}
