#include "defines.h"
#include "global.h"

#include <string>

using namespace std;

SDL_Surface* screen = NULL;
SDL_Surface* screenBuffer = NULL;
SDL_Surface* screenSecondaryBuffer = NULL;
volatile bool updateScreen = false;
volatile unsigned int cursorShape = CURSOR_ARROW;

volatile int mouseX, mouseY, mouseLB, mouseRB, key, ukey, kdir, tick;
volatile bool isActive = true;

bool updateAll = false;
list<SDL_Rect> updateList;

GameSettings settings, userSettings;
int enableSound = 1;

void PollEvents(void)
{
	SDL_Event event;

	while (SDL_PollEvent(&event))
	{
		switch (event.type) 
		{
			case SDL_KEYDOWN:
				key = event.key.keysym.sym;
				ukey = event.key.keysym.unicode;

				switch (key)
				{
					case SDLK_RIGHT: kdir |= 1; break;
					case SDLK_DOWN: kdir |= 2; break;
					case SDLK_LEFT: kdir |= 4; break;
					case SDLK_UP: kdir |= 8; break;
				}
				break;

			case SDL_KEYUP:
				key = 0;
				ukey = 0;

				switch (event.key.keysym.sym)
				{
					case SDLK_RIGHT: kdir &= 15^1; break;
					case SDLK_DOWN: kdir &= 15^2; break;
					case SDLK_LEFT: kdir &= 15^4; break;
					case SDLK_UP: kdir &= 15^8; break;
				}
				break;

			case SDL_MOUSEMOTION:
				mouseX = event.motion.x;
				mouseY = event.motion.y;
				break;

			case SDL_MOUSEBUTTONDOWN:
				if (event.button.button == SDL_BUTTON_LEFT) mouseLB = 1;
				else
				if (event.button.button == SDL_BUTTON_RIGHT) mouseRB = 1;
				break;

			case SDL_MOUSEBUTTONUP:
				if (event.button.button == SDL_BUTTON_LEFT) mouseLB = 0;
				else
				if (event.button.button == SDL_BUTTON_RIGHT) mouseRB = 0;
				break;

			case SDL_QUIT:
				isActive = false;
				break;
		}
	}

	tick = SDL_GetTicks();
}

void KeyPressed(void)
{
	key = 0;
	ukey = 0;
}

SDL_Rect DefRect(int x, int y, int w, int h)
{
	SDL_Rect r;
	r.x = x;
	r.y = y;
	r.w = w;
	r.h = h;
	return r;
}

void Clear(int c)
{
	int sz = screen->h * (screen->pitch >> 2);
	int* p = (int *)screen->pixels;

	while (sz--) *(p++) = c;
	updateList.push_back(DefRect(0, 0, screen->w, screen->h));
}

SDL_Surface *CopySurface(SDL_Surface *src)
{
	SDL_Rect s, d;
	SDL_PixelFormat *fmt;
	SDL_Surface *dst;

	fmt = screen->format;

	s.x = 0;
	s.y = 0;
	s.w = src->w;
	s.h = src->h;

	d.x = 0;
	d.y = 0;

	dst = SDL_CreateRGBSurface(SDL_SWSURFACE, src->w, src->h, fmt->BitsPerPixel, fmt->Rmask, fmt->Gmask, fmt->Bmask, 0);
	SDL_BlitSurface(src, &s, dst, &d);
	SDL_SetColorKey(dst, SDL_SRCCOLORKEY, DRGB(255, 0, 255));

	return dst;
}

SDL_Surface* LoadBitmap(const char* path)
{
	SDL_Surface* temp = NULL;
	temp = SDL_LoadBMP(path);
	if (!temp) Error((string("Error loaging bitmap ") + string(path)).c_str());

	SDL_Surface* conv = SDL_ConvertSurface(temp, screen->format, SDL_SWSURFACE);
	SDL_FreeSurface(temp);

	SDL_Surface* res = CopySurface(conv);
	SDL_FreeSurface(conv);

	return res;
}

void UpdateScreen(void)
{
	if (updateAll)
	{
		updateList.clear();
		updateList.push_back(DefRect(0, 0, screen->w, screen->h));
		updateAll = false;
	}

	updateScreen = true;
	while (updateScreen) SDL_Delay(1);
}

bool FrameBegin(void)
{
	if (SDL_MUSTLOCK(screen)) {
		if (SDL_LockSurface(screen) < 0) return false;
	}

	return true;
}

void FrameEnd(bool useDelay)
{
	if (SDL_MUSTLOCK(screen)) SDL_UnlockSurface(screen);

	UpdateScreen();
	if (useDelay) SDL_Delay(1);
}
