#include "font.h"
#include "exception.h"

char Font::availSymbols[] = "!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~";

Font::Font()
{
	for (int i = 0; i < 0x100; i++)
	{
		symbols[i].x = 0;
		symbols[i].y = 0;
		symbols[i].w = 0;
		symbols[i].h = 0;
	}

	alreadyInit = false;
}

Font::~Font()
{
}

void Font::Load(const char* path)
{
	fontPic.Load(path);
	InitSymbols();
}

void Font::Blit(SDL_Surface* dest, int x, int y, char c)
{
	if (c == '') c = '';

	SDL_Rect* s = &symbols[(unsigned int)(unsigned char)c];
	if (s->w==0 || s->h==0) return;

	SDL_Rect d;
	d.x = x;
	d.y = y;

	fontPic.Blit(dest, s, &d);
}

void Font::Blit(Picture& dest, int x, int y, char c)
{
	if (c == '') c = '';

	SDL_Rect* s = &symbols[(unsigned int)(unsigned char)c];
	if (s->w==0 || s->h==0) return;

	SDL_Rect d;
	d.x = x;
	d.y = y;

	fontPic.Blit(dest, s, &d);
}

int Font::Height(void)
{
	return fontPic.Height();
}

int Font::Width(char c)
{
	if (c == '') c = '';
	return symbols[(unsigned int)(unsigned char)c].w;
}

void Font::SetColor(int c)
{
	if (!alreadyInit) throw Exception("Load font before calling SetColor");

	for (int i = 0; i < 0x100; i++)
	{
		SDL_Rect* s = &symbols[i];
		if (s->w==0 || s->h==0) continue;

		int ex = s->x + s->w;
		int ey = s->y + s->h;

		for (int x = s->x; x < ex; x++)
		{
			for (int y = s->y; y < ey; y++)
			{
				int a = GET_A(fontPic.Pixel(x, y));

				if (a != 0) {
					fontPic.Pixel(x, y) = DRGBA(GET_R(c), GET_G(c), GET_B(c), a);
				}
			}
		}
	}
}

void Font::InitSymbols(void)
{
	int x = 0;
	int cnt = 0;

	for (char *p = availSymbols; *p; p++)
	{
		int n = (unsigned char)(*p);

		symbols[n].y = 0;
		symbols[n].h = fontPic.Height();

		x = GetLBound(x);
		int rx = GetRedLine(x);
		int ex = GetRBound(x, rx - 1);

		symbols[n].x = x;
		symbols[n].w = ex - x + 1;

		x = rx + 1;
		cnt++;
	}

	alreadyInit = true;
}

bool Font::CheckNoTransp(int x)
{
	for (int i = 0; i <= fontPic.Height(); i++) {
		if (GET_A(fontPic.Pixel(x, i))) return true;
	}

	return false;
}

bool Font::CheckRed(int x)
{
	for (int i = 0; i <= fontPic.Height(); i++) {
		if (GET_RGB(fontPic.Pixel(x, i)) == 0xFF0000) return true;
	}

	return false;
}

int Font::GetLBound(int x)
{
	while (x < fontPic.Width()) {
		if (CheckNoTransp(x)) break;
		x++;
	}

	return x;
}

int Font::GetRedLine(int x)
{
	while (x < fontPic.Width()) {
		if (CheckRed(x)) break;
		x++;
	}

	return x;
}

int Font::GetRBound(int lx, int x)
{
	while (x > lx) {
		if (CheckNoTransp(x)) break;
		x--;
	}

	return x;
}
