#include "pngloader.h"
#include "exception.h"
#include "png.h"

void ReadPngData(png_structp pngPtr, png_bytep data, png_size_t length)
{
	File* fl = (File *)png_get_io_ptr(pngPtr);
	unsigned int readed = fl->ReadBytes(data, length);

	if (readed != length) throw Exception("Error reading PNG");
}

void PrintPngError(png_structp pngPtr, png_const_charp msg)
{
	Error("PNG Error: %s", msg);
}

void PrintPngWarning(png_structp pngPtr, png_const_charp msg)
{
	AddLog("PNG Warning: %s", msg);
}

void LoadPNG(File* fl, int* width, int* height, DWORD** data, bool ignoreAlpha)
{
	png_byte buf[8];

	if (fl->ReadBytes(buf, 8) != 8) throw Exception("PNG header missing");
	if (!png_check_sig(buf, 8)) throw Exception("Invalid PNG header");

	png_struct* pngPtr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, PrintPngError, PrintPngWarning);
	if (!pngPtr) Error("png_create_read_struct : out of memory");

	png_info* infoPtr = png_create_info_struct(pngPtr);

	if (!infoPtr)
	{
		png_destroy_read_struct(&pngPtr, NULL, NULL);
		Error("png_create_info_struct : out of memory");
	}

	if (setjmp(png_jmpbuf(pngPtr)))
	{
		png_destroy_read_struct(&pngPtr, &infoPtr, NULL);
		Error("setjmp(png_jmpbuf(pngPtr)) : unknown error");
	}

	png_set_read_fn(pngPtr, fl, ReadPngData);
	png_set_sig_bytes(pngPtr, 8);
	png_read_info(pngPtr, infoPtr);

	png_uint_32 imageWidth;
	png_uint_32 imageHeight;
	int bitDepth;
	int colorType;
	double gamma;

	png_get_IHDR(pngPtr, infoPtr, &imageWidth, &imageHeight, &bitDepth, &colorType, NULL, NULL, NULL);

	if (bitDepth == 16) png_set_strip_16(pngPtr);
	if (colorType == PNG_COLOR_TYPE_PALETTE) png_set_expand(pngPtr);
	if (bitDepth < 8) png_set_expand(pngPtr);
	if (png_get_valid(pngPtr, infoPtr, PNG_INFO_tRNS)) png_set_expand(pngPtr);
	if (colorType==PNG_COLOR_TYPE_GRAY || colorType==PNG_COLOR_TYPE_GRAY_ALPHA) png_set_gray_to_rgb(pngPtr);
	if (png_get_gAMA(pngPtr, infoPtr, &gamma)) png_set_gamma(pngPtr, (double) 2.2, gamma);

	png_read_update_info(pngPtr, infoPtr); 
	png_get_IHDR(pngPtr, infoPtr, &imageWidth, &imageHeight, &bitDepth, &colorType, NULL, NULL, NULL);

	png_uint_32 rowBytes = png_get_rowbytes(pngPtr, infoPtr);	// row_bytes is the width x number of channels 
	png_uint_32 channels = png_get_channels(pngPtr, infoPtr);

	if (channels==3 || channels==4)
	{
		*width = imageWidth;
		*height = imageHeight;
		*data = new DWORD[imageWidth * imageHeight];

		png_bytep* rowPointers = new png_bytep[imageHeight];
		for (int i = 0; i < imageHeight; i++) rowPointers[i] = (BYTE*)&((*data)[i * imageWidth]);

		png_read_image(pngPtr, rowPointers);
		png_read_end(pngPtr, NULL);		// not really needed

		delete[] rowPointers;

		if (channels == 3)
		{
			for (int i = 0; i < imageHeight; i++)
			{
				BYTE* lineF = (BYTE*)&((*data)[i * imageWidth]) + rowBytes - 1;
				DWORD* lineT = &((*data)[i * imageWidth]) + imageWidth - 1;

				for (int j = imageWidth; j--;)
				{
					BYTE b = *(lineF--);
					BYTE g = *(lineF--);
					BYTE r = *(lineF--);

					*(lineT--) = DRGBA(r, g, b, 255);
				}
			}
		}
		else
		{
			for (int i = 0; i < imageHeight; i++)
			{
				DWORD* lineT = &((*data)[i * imageWidth]);
				BYTE* lineF = (BYTE*)lineT;

				for (int j = imageWidth; j--;)
				{
					BYTE r = *(lineF++);
					BYTE g = *(lineF++);
					BYTE b = *(lineF++);
					BYTE a = *(lineF++);

					*(lineT++) = DRGBA(r, g, b, (ignoreAlpha ? 255 : a));
				}
			}
		}
	}
	else { throw Exception("[PNG] Unsupported number of channels : %d", channels); }

	png_destroy_read_struct(&pngPtr, &infoPtr, NULL);
}
