#include "rectsfield.h"
#include "defines.h"

RectsField::RectsField()
{
}

RectsField::~RectsField()
{
}

void RectsField::Clear()
{
	rects.clear();
}

void RectsField::AddRect(SDL_Rect& rect)
{
	wrects.clear();
	wrects.push_back(rect);
	Intersect();
}

void RectsField::AddRect(SDL_Rect* rect)
{
	wrects.clear();
	wrects.push_back(*rect);
	Intersect();
}

void RectsField::BoundRect(SDL_Rect& rect)
{
	BoundIntersect(&rect);
}

void RectsField::BoundRect(SDL_Rect* rect)
{
	BoundIntersect(rect);
}

SDL_Rect RectsField::MkRect(int x, int y, int w, int h)
{
	SDL_Rect r;
	r.x = x;
	r.y = y;
	r.w = w;
	r.h = h;
	return r;
}

void RectsField::AddWRect(SDL_Rect r)
{
	if (r.h>0 && r.w>0) wrects.push_back(r);
}

bool RectsField::IntersectRects(SDL_Rect* a, SDL_Rect* b)
{
	int aex = a->x + a->w;
	int aey = a->y + a->h;
	int bex = b->x + b->w;
	int bey = b->y + b->h;

	/*	....
		.AA.
		.AA.
		.... */
	if (a->x>=b->x && a->y>=b->y && aex<=bex && aey<=bey) return false;

	/* Not intersect */
	if (a->x>=bex || aex<b->x || a->y>=bey || aey<b->y) return true;

	/*	AAAA
		A..A
		A..A
		AAAA */
	if (b->x>=a->x && b->y>=a->y && bex<=aex && bey<=aey)
	{
		AddWRect(MkRect(bex, a->y, aex - bex, a->h));
		AddWRect(MkRect(b->x, a->y, b->w, b->y - a->y));
		AddWRect(MkRect(b->x, bey, b->w, aey - bey));

		if (a->x != b->x)
		{
			a->w = b->x - a->x;
			return true;
		}

		return false;
	}

	if (a->x<=b->x && aex>=bex)
	{
		/*	AAAA
			A..A
			 ..  */
		if (a->y<=b->y && aey<=bey)
		{
			b->y = aey;
			b->h = bey - aey;
			return true;
		}

		/*	 ..
			A..A
			A..A
			 ..  */
		if (a->y>b->y && aey<bey)
		{
			AddWRect(MkRect(bex, a->y, aex - bex, a->h));
			a->w = b->x - a->x;
			return true;
		}

		/*	 ..
			A..A
			AAAA */
		b->h = a->y - b->y;
		return true;
	}

	if (a->y<=b->y && aey>=bey)
	{
		/*	AA
			A::
			AA  */
		if (a->x<=b->x && aex<=bex)
		{
			b->x = aex;
			b->w = bex - aex;
			return true;
		}

		/*	 A
			:::
			 A  */
		if (a->x>b->x && aex<bex)
		{
			AddWRect(MkRect(a->x, bey, a->w, aey - bey));
			a->h = b->y - a->y;
			return true;
		}

		/*	 AA
			::A
			 AA */
		b->w = a->x - b->x;
		return true;
	}

	if (a->x<b->x && aex<=bex)
	{
		/*	AA
			A..
			 .. */
		if (a->y<b->y && aey<=bey)
		{
			AddWRect(MkRect(b->x, a->y, aex - b->x, b->y - a->y));
			a->w = b->x - a->x;
			return true;
		}

		/*	 ..
			AA.
			 .. */
		if (a->y>=b->y && aey<=bey)
		{
			a->w = b->x - a->x;
			return true;
		}

		/*	 ..
			A..
			AA  */
		AddWRect(MkRect(b->x, bey, aex - b->x, aey - bey));
		a->w = b->x - a->x;
		return true;
	}

	if (a->x>=b->x && aex<=bex)
	{
		/*	 AA
			.AA.
			.... */
		if (a->y < b->y)
		{
			a->h = b->y - a->y;
			return true;
		}

		/*	....
			.AA. <- can't be there
			.AA.
			.... */

		/*	....
			.AA.
			 AA  */
		a->y = bey;
		a->h = aey - bey;
		return true;
	}

	/*	 AA
		..A
		..  */
	if (a->y<b->y && aey<=bey)
	{
		AddWRect(MkRect(bex, a->y, aex - bex, a->h));
		a->w = bex - a->x;
		a->h = b->y - a->y;
		return true;
	}

	/*	..
		.AA
		..  */
	if (a->y>=b->y && aey<=bey)
	{
		a->x = bex;
		a->w = aex - bex;
		return true;
	}

	/*	..
		..A
		 AA */
	AddWRect(MkRect(a->x, bey, bex - a->x, aey - bey));
	a->x = bex;
	a->w = aex - bex;
	return true;
}

void RectsField::Intersect(void)
{
	for (int k = 0; k < wrects.size(); k++)
	{
		int i;

		int cnt = rects.size();
		SDL_Rect rc = wrects[k];

		for (i = 0; i < cnt; i++) {
			if (!IntersectRects(&rc, &rects[i])) break;
		}

		if (i >= cnt) rects.push_back(rc);
	}
}

bool RectsField::IntersectRectsBound(SDL_Rect *a, SDL_Rect *b)
{
	int aex = a->x + a->w;
	int aey = a->y + a->h;
	int bex = b->x + b->w;
	int bey = b->y + b->h;

	/*	....
		.AA.
		.AA.
		.... */
	if (a->x>=b->x && a->y>=b->y && aex<=bex && aey<=bey) return true;

	/* Out of bounding rect */
	if (a->x>=bex || aex<=b->x || a->y>=bey || aey<=b->y) return false;

	/*	AAAA
		A..A
		A..A
		AAAA */
	if (b->x>=a->x && b->y>=a->y && bex<=aex && bey<=aey)
	{
		a->x = b->x;
		a->y = b->y;
		a->w = b->w;
		a->h = b->h;
		return true;
	}

	if (a->x<b->x && aex>bex)
	{
		/*	AAAA
			A..A
			 ..  */
		if (a->y<b->y && aey<=bey)
		{
			a->x = b->x;
			a->y = b->y;
			a->w = b->w;
			a->h = aey - a->y;
			return true;
		}

		/*	 ..
			AAAA
			AAAA
			 ..  */
		if (a->y>=b->y && aey<=bey)
		{
			a->x = b->x;
			a->w = b->w;
			return true;
		}

		/*	 ..
			A..A
			AAAA */
		a->x = b->x;
		a->w = b->w;
		a->h = bey - a->y;
		return true;
	}

	if (a->y<b->y && aey>bey)
	{
		/*	AA
			A::
			AA  */
		if (a->x<b->x && aex<=bex)
		{
			a->x = b->x;
			a->y = b->y;
			a->w = aex - a->x;
			a->h = b->h;
			return true;
		}

		/*	 AA
			:AA:
			 AA  */
		if (a->x>=b->x && aex<=bex)
		{
			a->y = b->y;
			a->h = b->h;
			return true;
		}

		/*	 AA
			::A
			 AA */
		a->y = b->y;
		a->w = bex - a->x;
		a->h = b->h;
		return true;
	}

	if (a->x<b->x && aex<=bex)
	{
		/*	AA
			A..
			 .. */
		if (a->y<b->y && aey<=bey)
		{
			a->x = b->x;
			a->y = b->y;
			a->w = aex - a->x;
			a->h = aey - a->y;
			return true;
		}

		/*	 ..
			AA.
			AA.
			 .. */
		if (a->y>=b->y && aey<=bey)
		{
			a->x = b->x;
			a->w = aex - a->x;
			return true;
		}

		/*	 ..
			A..
			AA  */
		a->x = b->x;
		a->w = aex - a->x;
		a->h = bey - a->y;
		return true;
	}

	if (a->x>=b->x && aex<=bex)
	{
		/*	 AA
			.AA.
			.... */
		if (a->y < b->y)
		{
			a->y = b->y;
			a->h = aey - a->y;
			return true;
		}

		/*	....
			.AA. <- can't be there
			.AA.
			.... */

		/*	....
			.AA.
			 AA  */
		a->h = bey - a->y;
		return true;
	}

	/*	 AA
		..A
		..  */
	if (a->y<b->y && aey<=bey)
	{
		a->y = b->y;
		a->w = bey - a->x;
		a->h = aey - a->y;
		return true;
	}

	/*	..
		.AA
		.AA
		..  */
	if (a->y>=b->y && aey<=bey)
	{
		a->w = bex - a->x;
		return true;
	}

	/*	..
		..A
		 AA */
	a->w = bex - a->x;
	a->h = bey - a->y;
	return true;
}

void RectsField::BoundIntersect(SDL_Rect* bnd)
{
	brects.clear();
	vector<SDL_Rect>::iterator ri;

	for (ri = rects.begin(); ri != rects.end(); ri++)
	{
		SDL_Rect rc = *ri;
		if (IntersectRectsBound(&rc, bnd)) brects.push_back(rc);
	}

	rects.clear();
	list<SDL_Rect>::iterator it;

	for (it = brects.begin(); it != brects.end(); it++) rects.push_back(*it);
}
