#include "settings.h"
#include "file.h"

GameSettings::GameSettings()
{
}

GameSettings::~GameSettings()
{
}

void GameSettings::SkipSpaces(void)
{
	while (*ptr && *ptr==' ') ptr++;
}

bool GameSettings::IsValidSym(char c)
{
	return ((c>='0' && c<='9') || (c>='a' && c<='z') || (c>='A' && c<='Z') || c=='_' || c=='.');
}

string GameSettings::ReadName(void)
{
	string res("");
	if (!(*ptr && IsValidSym(*ptr))) return res;

	do {
		res += *(ptr++);
	} while ((*ptr) && IsValidSym(*ptr));

	return res;
}

string GameSettings::ReadPrefix(void)
{
	string res("");
	if (!(*ptr && IsValidSym(*ptr))) return res;

	do {
		res += *(ptr++);
	} while ((*ptr) && IsValidSym(*ptr));

	return res;
}

bool GameSettings::IsNum(char c)
{
	return (c>='0' && c<='9');
}

bool GameSettings::IsHexNum(char c)
{
	return ((c>='0' && c<='9') || (c>='a' && c<='f') || (c>='A' && c<='F'));
}

int GameSettings::ReadNumber(void)
{
	int res = 0;
	if (!(*ptr)) return res;

	if (*ptr == '-')
	{
		ptr++;

		if (!(*ptr && IsNum(*ptr)))
		{
			ptr--;
			return res;
		}

		if (*ptr == '0' && ptr[1]=='x')
		{
			ptr += 2;

			while (*ptr && IsHexNum(*ptr))
			{
				res = 16 * res + unhex(*ptr);
				ptr++;
			}
		}
		else
		{
			while (*ptr && IsNum(*ptr))
			{
				res = 10 * res + (*ptr - '0');
				ptr++;
			}
		}

		return -res;
	}
	else
	{
		if (*ptr == '0' && ptr[1]=='x')
		{
			ptr += 2;

			while (*ptr && IsHexNum(*ptr))
			{
				res = 16 * res + unhex(*ptr);
				ptr++;
			}
		}
		else
		{
			while (*ptr && IsNum(*ptr))
			{
				res = 10 * res + (*ptr - '0');
				ptr++;
			}
		}

		return res;
	}
}

void GameSettings::Load(const char* path)
{
	File fl;
	fl.Read(path);

	string prefix("");

	while (!fl.Eof())
	{
		fl.GetS(buffer, 0x1000);

		ptr = buffer;
		SkipSpaces();

		if (!(*ptr)) continue;

		if (*ptr == '[')
		{
			ptr++;
			SkipSpaces();
			prefix = ReadPrefix();
		}
		else
		{
			string name = ReadName();
			if (name.empty()) continue;

			SkipSpaces();
			if (*ptr != '=') continue;

			ptr++;
			SkipSpaces();

			char *bptr = ptr;
			int val = ReadNumber();

			if (*ptr || ptr==bptr) continue;

			if (prefix.empty()) values[name] = val;
			else values[prefix + string(".") + name] = val;
		}
	}

	fl.Close();
}

int GameSettings::Get(const char *name, int defaultValue)
{
	string nm(name);
	if (values.find(nm) == values.end()) return defaultValue;
	return values[nm];
}
