#include "sound.h"
#include "SDL.h"

#define SOUND_BUFFER_SIZE 0x1000

MusicChannel channels[MAX_CHANNELS];

void AudioCallback(void *userData, Uint8 *stream, int len)
{
	static BYTE buffer[SOUND_BUFFER_SIZE];
	memset(stream, 0, len);

	for (int i = 0; i < MAX_CHANNELS; i++)
	{
		if (channels[i].IsLoaded())
		{
			channels[i].FillBuffer(buffer, len);

			BYTE* o = buffer;
			BYTE* p = stream;
			unsigned int sz = len / 2;

			while (sz--)
			{
				int f = (Sint16)((o[1] << 8) | (*o));
				int t = (Sint16)((p[1] << 8) | (*p));

				t += f;
				if (t > 32767) t = 32767;
				if (t < -32768) t = -32768;

				*(p++) = (t & 0xFF);
				*(p++) = (t >> 8);

				o += 2;
			}
		}
	}
}

void InitAudio(void)
{
	SDL_AudioSpec desired, obtained;

	desired.freq = 44100;
	desired.format = AUDIO_S16LSB;
	desired.channels = 2;
	desired.samples = SOUND_BUFFER_SIZE / desired.channels / 2 /* 16 bit */;
	desired.callback = AudioCallback;

	if (SDL_OpenAudio(&desired, &obtained) < 0)
	{
		SDL_Quit();
		Error("[SDLWavePlayer::Init] Couldn't open audio: %s\n", SDL_GetError());
	}

	SDL_PauseAudio(0);
}
